// Tutorial for VDShader video filter.
//
// To use VDShader, you must create a shader. This shader is very similar to the pixel shaders
// normally used with 3D graphics, and tells the processing engine how to fetch and process
// pixels in the video. Shaders are written in a C-like language that is adapted to parallel
// processing of vectors instead of scalars, so that instead of having expressions that process
// individual values, expressions process all of the components of a pixel in parallel instead.
//
// For those of you familiar with HLSL/Cg, what appears below will look familiar. In fact, it's
// a perfectly valid effect file that will compile with fxc. However, to actually be used for
// hardware rendering, it'd need a vertex shader and a texture or annotations on the sampler,
// which VDShader doesn't support right now. The vertex shader is basically supplied for you
// instead.

// Anyway, to start, we must declare where our pixels will come from:

sampler src;

// A 'sampler' fetches pixels from a source, which in this case is the incoming video stream.

// Next, we will declare the pixel shader function.
//
// float4: The pixel shader returns a vector with four numbers, corresponding to red, green,
//         blue, and alpha (ignored).
//
// MyPixelShader: The name of the pixel shader. Note that this shouldn't be called PixelShader.
//
// float2 uv: The parameter to a PixelShader is a 2-vector with the (x,y) coordinate of the
//            pixel we're processing. Both values go from 0 to 1.
//
// TEXCOORD0: This is known as a semantic, and tells the compiler that 'uv' should be set to
//            the coordinate of the current pixel.
//
// COLOR0: This is also a semantic, and tells the compiler that the float4 return value is
//         a color output.
//
// uniform float saturation:
//    This declares a uniform (non-changing) parameter for the pixel shader. These are
//    optional, and you don't need to include any, but they're useful for customizing a shader
//    that is called multiple times. More on this later.

float4 MyPixelShader(float2 uv : TEXCOORD0, uniform float saturation) : COLOR0 {
	// Now we must declare the body of the pixel shader. Let's start by declaring a temporary
	// variable:
	float4 c;
	
	// Next, let's fetch the pixel that we're supposed to be processing, and store it into
	// the variable. We use our one sampler as the source, and the coordinates we received
	// as the location. The result is that we get a 4-vector with red, green, blue, and
	// alpha from 0-1. The alpha channel is unused.
	c = tex2D(src, uv);

	// Now let's do something with the pixel. How about increasing the saturation and
	// contrast? First, we must separate brightness and color information. We can do this
	// by computing Y = 0.30R + 0.59G + 0.11B. This is easily done via a 'dot product.'
	
	// We declare a vector with three constants indicating the contributions of red, green,
	// and blue to the luma output.
	float3 lumaAxis = float3(0.30, 0.59, 0.11);
	
	// Now we dot it against the red, green, and blue channels. The .rgb part is known as a
	// swizzle, and extracts channels in a certain order from an expression. Here we're just
	// isolating red/green/blue, but it's also possible to extract different sizes and
	// orders:
	//
	//	.r = red
	//	.rb = float2(red, blue)
	//	.gbg = float3(green, blue, green)
	//	.rgba = float4(red, green, blue, alpha) (no change)
	//
	// x/y/z/w can also be used as a synonym for r/g/b/a.
	//
	// dot(x,y) is known as an intrinsic function, and on float3 inputs returns (x.r*y.r +
	// x.g*y.g + x.b*y.b). There are also variants for float2 and float4. Other available
	// intrinsics are length(), to return the length of a vector; normalize(), which changes
	// the length of a vector to 1 in the same direction, and mul(), which multiplies a
	// vector by a matrix.
	//
	float y = dot(c.rgb, lumaAxis);
	
	// Now that we have luma, we can subtract it from the original pixel to get chroma.
	// Because this is a vector language, regular operators are extended to parallel
	// processing. When you add two vectors via (a+b), you're actually adding all four
	// components in parallel, so the result is a.r+b.r, a.g+b.g, etc. The same is true
	// of -, *, and /. In general, it's a lot easier to process pixels this way, because
	// you spend less time splitting pixels apart into red/green/blue components.
	//
	// Notice that we're subtracting a scalar (float), which only has one component, from
	// a float4, which has four. The compiler automatically widens shorter vectors as needed
	// to match them. Scalars get expanded via replication, so in this case, we're
	// computing float4(px.r - y, px.g - y, px.b - y, px.a - y). Longer vectors are padded
	// with zeroes instead. If we wanted to add brightness here, we could do it just by
	// adding the brightness value, i.e. (c - y) + 0.1.
	c = c - y;
	
	// Finally, let's increase saturation by amplifying the chroma, and combine it back
	// with luma. But what about the alpha channel? It's unused in this filter. However,
	// let's say we're a stickler, and want to set it to 1 for some reason. Well, we can
	// do that by constructing a new float4 with 1 in the alpha channel. Because this
	// language is meant for processing vectors, we don't need to mess around with passing
	// .r, .g, and .b -- we just pass a float3 and a float, and the result is a float4.
	return float4((c * saturation).rgb + y, 1);
}

// Our pixel shader is now defined, so all that's left is to tell the compiler to use it.
// We do this by declaring a technique. Techniques are the top-level programs that are
// used for rendering. VDShader will compile multiple techniques, but right now it will
// only use the first one. The technique is run on each frame individually, over and over.
//
// The name on the technique is optional, but it never hurts to have one.

technique MyTechnique {
	// Next, we must declare a pass. Each pass is a run over the image; multiple passes
	// can be used to composite. Again, however, VDShader only supports the first pass
	// right now.
	pass {
		// Finally, we select the pixel shader. The 'compile' statement here actually
		// compiles the pixel shader. The 1.5 argument that we supply here is put
		// into the uniform 'saturation' parameter we declared above, and the pixel
		// shader is compiled with that value as a constant. The 'uv' parameter doesn't
		// get specified here because it's supplied by the runtime.
		//
		// The 'ps_2_0' is here for compatibility with other compilers. It normally
		// specifies the pixel shader profile that the compiler targets, and determines
		// what level of hardware is necessary to run the shader. VDShader doesn't care.
		
		PixelShader = compile ps_2_0 MyPixelShader(1.5);
	}
}
